import time
from collections.abc import Callable

import pytest
from delayedqueue.delayed_queue_item import DelayedQueueItem


@pytest.fixture
def time_func():
    return time.time


@pytest.fixture
def delayed_queue_item(time_func):
    return DelayedQueueItem(item="an item", delay=10, time_func=time_func)


@pytest.fixture
def another_delayed_queue_item(time_func):
    return DelayedQueueItem(item="another item", delay=10, time_func=time_func)


def test_delayed_queue_item(time_func, delayed_queue_item, another_delayed_queue_item):
    assert delayed_queue_item.item == "an item"
    assert delayed_queue_item.delay == 10
    assert isinstance(time_func, Callable)

    assert delayed_queue_item != another_delayed_queue_item
    assert delayed_queue_item < another_delayed_queue_item


def test_delayed_queue_item_raise_errs(delayed_queue_item, another_delayed_queue_item):
    with pytest.raises(ValueError):
        assert delayed_queue_item == 10

    with pytest.raises(ValueError):
        assert another_delayed_queue_item < 10


def test_delayed_queue_item_delay(delayed_queue_item):
    item_delay = delayed_queue_item.get_delay()
    assert type(item_delay) == float
    assert (delayed_queue_item.target_time - int(delayed_queue_item.time_func() * 1000)) / 1000 == item_delay



import time

import pytest

from delayedqueue import DelayedQueue


@pytest.fixture
def delayedqueue():
    return DelayedQueue()


@pytest.fixture
def custom_time_func():
    return lambda: 1000


def test_delayedqueue_when_empty(delayedqueue):
    assert delayedqueue.is_empty() == True
    assert delayedqueue.peek() is None


def test_delayedqueue_ops(delayedqueue):
    assert delayedqueue.put("an item", 0) == True
    assert delayedqueue.peek() == "an item"
    assert delayedqueue.get() == "an item"

    with pytest.raises(ValueError):
        assert delayedqueue.put("test item", -1)


def test_delayedqueue_time_func(capfd, delayedqueue, custom_time_func):
    assert delayedqueue.time_func == time.time
    delayedqueue.time_func = custom_time_func
    assert delayedqueue.time_func() == 1000

    with pytest.raises(ValueError):
        delayedqueue.time_func = "definitely not a callable"

    delayedqueue.put("an item", 0)
    delayedqueue.time_func = custom_time_func
    out, err = capfd.readouterr()
    assert out.strip() == "Warning! updating time_func after enqueue. May result in unknown behaviour!"




import threading
import time
from collections.abc import Callable
from typing import Callable as _Callable
from heapq import heappush, heappop

from delayedqueue.delayed_queue_item import DelayedQueueItem


class DelayedQueue:

    def __init__(self):
        self.queue = []
        self.lock = threading.Lock()
        self.available = threading.Condition(self.lock)
        self.__time_func = time.time

    @property
    def time_func(self) -> _Callable:
        return self.__time_func

    @time_func.setter
    def time_func(self, func):
        if not isinstance(func, Callable):
            raise ValueError(f"time_func setter value [{func}] should be a callable")
        if self.queue:
            print("Warning! updating time_func after enqueue. May result in unknown behaviour!")
        self.__time_func = func

    def is_empty(self) -> bool:
        return len(self.queue) == 0

    def peek(self):
        with self.lock:
            if self.queue:
                return self.queue[0].item

    def put(self, item, delay: int) -> bool:
        if delay < 0:
            raise ValueError("delay must be a non-negative number")

        with self.available:
            wrapped_item = DelayedQueueItem(item, delay, self.__time_func)
            heappush(self.queue, wrapped_item)

            if self.queue[0] == wrapped_item:
                self.available.notify()
            return True

    def get(self):
        with self.lock:
            try:
                while True:
                    if not self.queue:
                        self.available.wait()
                    first = self.queue[0]
                    delay = first.get_delay()
                    if delay <= 0:
                        wrapped_item = heappop(self.queue)
                        return wrapped_item.item

                    self.available.wait(delay)
            finally:
                if self.queue:
                    self.available.notify()



"""
Wrapper for the item which is going to be enqueued in the queue.
"""


class DelayedQueueItem:
    def __init__(self, item, delay, time_func):
        self.item = item
        self.delay = delay
        self.time_func = time_func
        self.target_time = int(self.time_func() * 1000) + delay * 1000

    def get_delay(self) -> int:
        return (self.target_time - int(self.time_func() * 1000)) / 1000

    def __lt__(self, other) -> bool:
        if not isinstance(other, DelayedQueueItem):
            raise ValueError("comparison should be made between same types")
        return self.target_time <= other.target_time

    def __eq__(self, other) -> bool:
        if not isinstance(other, DelayedQueueItem):
            raise ValueError("comparison should be made between same types")
        return self.item == other.item and self.delay == other.delay and self.target_time == other.target_time

    def __repr__(self) -> str:
        return f"""(Item: {self.item}, Delay: {self.delay}, Time Func: {self.time_func}, Next Execution Time: {self.target_time}"""
